import os

PROCESS_ERROR_MSG_TEMPLATE = "Table function process failed: function '{}', {}"
VALIDATE_ERROR_MSG_TEMPLATE = "parameters type should be [{}]"

STONEWAVE_HOME = os.environ.get("STONEWAVE_HOME", "/tmp")
USER_DEFINED_TABLE_FUNCTIONS_MODULE = "py_udtfs"
USER_DEFINED_TABLE_FUNCTIONS_PATH = os.path.join(STONEWAVE_HOME, "var", USER_DEFINED_TABLE_FUNCTIONS_MODULE)
USER_DEFINED_TABLE_FUNCTION_INFO_FILE = "info.toml"
SIGNATURE_LIST = "signature_list"

# for ci tests
if not os.path.exists(USER_DEFINED_TABLE_FUNCTIONS_PATH):
    os.makedirs(USER_DEFINED_TABLE_FUNCTIONS_PATH)


class ParameterType:
    LITERAL = "literal"
    COLUMN = "column"
    TYPE = "type"
    VALUE = "value"


class ParameterDataType:
    STRING = "STRING"
    INT = "INT"
    FLOAT = "FLOAT"
    BOOL = "BOOL"
    TABLE = "TABLE"

    def with_default_value(datatype, default_val):
        return "{} = {}".format(datatype, default_val)



import pyarrow as pa
from stonewave.sql.udtfs.logger import logger


class RecordBatchBuilder(object):
    def __init__(self):
        self.columns = {}
        self.schema = {}
        self._row_count = 0

    def column_names(self):
        return self.schema.keys()

    def add_column(self, column_name, data_type):
        """
        Append a column to the builder. If the column with the same name exists in the builder, it will NOT be added again.
        :param column_name: the name of the column
        :param data_type: pyarrow's field data type, for example, pyarrow::utf8()
        :return: None
        """
        if column_name not in self.columns:
            self.columns[column_name] = []
            self.schema[column_name] = data_type

    def append(self, column, value):
        """
        Append a new value for a given column.
        After appending values for all columns in a row,
        caller needs to call `increase_row_count` to ensure the builder's row_count is increased correctly.
        :param column: the name of the column
        :param value: the new value to be append to the column
        :return:
        """
        self._fill_null_for_all_builders(self._row_count)
        return self.columns.get(column).append(value)

    def extend(self, column, values):
        """
        Append multiple new values for a new given column.
        After appending values for all columns in a row,
        caller needs to call `increase_row_count` to ensure the builder's row_count is increased correctly.
        :param column: the name of the column
        :param values: a list of new values to be appended to the columns
        :return:
        """
        # caller needs to call `increase_row_count` to ensure the builder's row_count is increased correctly
        self._fill_null_for_all_builders(self._row_count)
        return self.columns.get(column).extend(values)

    def num_fields(self):
        """
        :return: the number of fields (columns) currently in this builder
        """
        return len(self.schema)

    def flush(self):
        """
        Flush the contents in the record build to build a record batch.
        return the record batch built from the contents in the builder,
        and return None if the built record batch is empty.
        """
        if self.columns:
            logger.debug(
                "flushing contents from record batch builder",
                columns=len(self.columns),
                row_count=self._row_count,
            )
            self._fill_null_for_all_builders(self._row_count)
            arrow_arrays = [pa.array(array, self.schema[col]) for col, array in self.columns.items()]
            batch = pa.RecordBatch.from_arrays(arrow_arrays, list(self.columns.keys()))
            # clear all arrays
            for col in self.columns.keys():
                self.columns[col] = []
            self._row_count = 0
            return batch
        # having null rows only, use empty batch to indicates this
        elif self._row_count > 0:
            self._row_count = 0
            batch = pa.RecordBatch.from_arrays([])
            return batch
        else:
            return None

    def row_count(self):
        """
        Return the number of rows in the builder
        :return:
        """
        return self._row_count

    def increase_row_count(self, increment=1):
        """
        Increase the number of rows in the builder. This is necessary after caller append values for individual column
        by calling `append` or `extend` API in this class.
        :param increment: the number of rows to be increased, by default, this method increases by 1
        :return:
        """
        self._row_count += increment

    def append_null_row(self):
        """
        append a null row to the builder.
        The row count will be automatically maintained.
        :return:
        """
        for array in self.columns.values():
            array.append(None)
        self.increase_row_count()

    def append_row(self, kv_pairs):
        """
        Append a row with as key value pairs, represented by a Python dictionary.
        The key represents the column name and the value represents the column value, which is pyarrow.utf8() type
        :param kv_pairs: the row's field names and values
        :return:
        """
        self._fill_null_for_all_builders(self._row_count)
        for key, value in kv_pairs.items():
            # TODO: we don't support creating multi value field yet in python
            self.add_column(key, pa.utf8())
            array = self._get_column_array(key)
            array.append(value)
        self.increase_row_count()

    def _fill_null_for_all_builders(self, new_row_count):
        for array in self.columns.values():
            null_count = new_row_count - len(array)
            for _ in range(0, null_count):
                array.append(None)

    def _get_column_array(self, column):
        return self.columns.get(column)



import pyarrow as pa
from pygrok import Grok
from stonewave.sql.udtfs.base_function import BaseFunction, udtf


@udtf(is_parser=True)
class ParseGrokFunction(BaseFunction):
    def __init__(self):
        self.grok_pattern_cache = {}

    def get_name(self):
        return "parse_grok"

    def process(self, row_writer, row_idx, args):
        text = args[0]
        grok_pattern = args[1]

        grok = None
        if grok_pattern in self.grok_pattern_cache:
            grok = self.grok_pattern_cache[grok_pattern]
        else:
            grok = Grok(grok_pattern)
            self.grok_pattern_cache[grok_pattern] = grok
        extracted_values = grok.match(text)

        record_batch_builder = row_writer.record_batch_builder
        if extracted_values:
            for col_name, col_value in extracted_values.items():
                # FIXME: support more non string data types
                record_batch_builder.add_column(col_name, pa.utf8())
                record_batch_builder.append(col_name, col_value)
            record_batch_builder.increase_row_count()



def rename_duplicated_field(schema, field):
    name = field.name
    field_id = 0
    while schema.get_field_index(name) != -1:
        field_id += 1
        name = "{}${}".format(field.name, field_id)
    return field if name == field.name else field.with_name(name)



import pyarrow as pa
import pandas as pd
import pathlib
from stonewave.sql.udtfs.logger import logger
from stonewave.sql.udtfs.base_function import BaseFunction


class LoadExcelFunction(BaseFunction):
    def __init__(self):
        pass

    def get_name(self):
        return "load_excel"

    def process(self, row_writer, row_idx, args):
        excel_file = args[0]
        sheet_name = [c.strip() for c in args[1].split(",")] if len(args) > 1 and args[1] != "" else None
        logger.debug(
            "executing load_excel table function",
            excel_file=excel_file,
            sheet_name=sheet_name,
        )
        try:
            excel_file_path = pathlib.Path(excel_file)
            if excel_file_path.exists():
                excel_data = pd.read_excel(
                    excel_file_path,
                    sheet_name=sheet_name,
                    engine="openpyxl",
                    dtype=str,
                )

                sheets = []
                for sheet_name, df in excel_data.items():
                    # trim spaces in column names
                    df.columns = df.columns.str.strip()
                    table = pa.Table.from_pandas(df, preserve_index=False)
                    batches = table.to_batches()
                    if batches:
                        sheets.append(batches[0])
                row_writer.batch_iterator = iter(sheets)
        except Exception as e:
            logger.error("failed to load excel file", error=str(e))



import sys
import json
from stonewave.sql.udtfs.load_function import load_function_by_name
from stonewave.sql.udtfs.logger import logger
from stonewave.sql.udtfs.constants import USER_DEFINED_TABLE_FUNCTIONS_PATH
from stonewave.sql.udtfs.protocol.fsm.apply_function_fsm import ApplyFunctionFsm
from stonewave.sql.udtfs.protocol.fsm.eval_function_fsm import EvalFunctionFsm
from stonewave.sql.udtfs.protocol.fsm.eval_function_with_table_param_fsm import (
    EvalFunctionWithTableParamFsm,
)
from stonewave.sql.udtfs.protocol.fsm.result_batch_sender import (
    SharedMemoryRecordBatchSender,
)


# this protocol roughly follows json-rpc protocol's request/response format
# stdin/stdout is used as transport (in_func/out_io)
# https://en.wikipedia.org/wiki/JSON-RPC
# a request contains three properties: method, params, id
# a response contains three properties: result, error, id
def execute(func_name, in_func, out_io):
    sys.path.append(USER_DEFINED_TABLE_FUNCTIONS_PATH)
    logger.info("start function execution", function=func_name)

    function = load_function_by_name(func_name)
    func = function()

    result = []
    fsm = None
    batch_sender = SharedMemoryRecordBatchSender()

    while True:
        logger.debug("waiting for request", function=func_name)
        command = in_func()
        logger.debug("receive request", function=func_name, request=command)
        request = json.loads(command)
        method, params, request_id = request["method"], request["params"], request["id"]

        def respond(result, error=None):
            res = json.dumps({"result": result, "error": error, "id": request_id})
            logger.debug("send response", function=func_name, response=res)
            print(
                res,
                file=out_io,
                flush=True,
            )

        if fsm is None:
            if method == "apply":
                fsm = ApplyFunctionFsm(func, batch_sender)
            elif method == "eval":
                fsm = EvalFunctionFsm(func, batch_sender)
            elif method == "eval_with_table_param":
                fsm = EvalFunctionWithTableParamFsm(func, batch_sender)
        fsm_trigger = getattr(fsm, method, None)
        if fsm_trigger:
            try:
                fsm_trigger(params, respond)
            except Exception as e:
                logger.error("error occurred during function execution", error=str(e))
                respond(result=None, error=e.args[0])
                break
            logger.debug("finish executing request", function=func_name, state=fsm.state)
            if fsm.is_end():
                logger.info("finish function execution", function=func_name)
                break
        else:
            respond(result=None, error="invalid_method=" + method)


if __name__ == "__main__":
    func_name = sys.argv[1]
    execute(func_name, input, sys.stdout)



import logging.config
import structlog
import os
import pathlib
from structlog.contextvars import (
    bind_contextvars,
    clear_contextvars,
    merge_contextvars,
    unbind_contextvars,
)

LOG_LEVEL = os.getenv("SW_LOG_LEVEL") or "INFO"

log_file_dir = os.path.join(os.getenv("STONEWAVE_HOME", "/tmp"), "var", "logs", "py_table_funcs")
pathlib.Path(log_file_dir).mkdir(parents=True, exist_ok=True)
log_file = os.path.join(log_file_dir, "py_table_funcs.{}.log".format(os.getpid() % 1000))

timestamper = structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S")
pre_chain = [
    # Add the log level and a timestamp to the event_dict if the log entry
    # is not from structlog.
    structlog.stdlib.add_log_level,
    timestamper,
]

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.dev.ConsoleRenderer(colors=False),
                "foreign_pre_chain": pre_chain,
            },
        },
        "handlers": {
            "file": {
                "level": LOG_LEVEL,
                "class": "logging.handlers.RotatingFileHandler",
                "filename": log_file,
                "formatter": "plain",
                "maxBytes": 2097152,  # 2MB
                "backupCount": 1,
            },
        },
        "loggers": {
            "": {
                "handlers": ["file"],
                "level": LOG_LEVEL,
                "propagate": True,
            },
        },
    }
)
structlog.configure(
    processors=[
        merge_contextvars,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        timestamper,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ],
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=True,
)

logger = structlog.get_logger()
bind_contextvars(pid=os.getpid())



import pyarrow as pa
from multiprocessing import shared_memory
from multiprocessing.resource_tracker import unregister


def _get_batch_data_size(record_batch):
    mock_sink = pa.MockOutputStream()
    stream_writer = pa.RecordBatchStreamWriter(mock_sink, record_batch.schema)
    stream_writer.write_batch(record_batch)
    stream_writer.close()
    data_size = mock_sink.size()
    return data_size


# FIXME: test arrow.ipc.writer to see if we can avoid copying
def _write_batch_to_buf(record_batch, buf):
    stream = pa.FixedSizeBufferWriter(buf)
    stream_writer = pa.RecordBatchStreamWriter(stream, record_batch.schema)
    stream_writer.write_batch(record_batch)
    stream_writer.close()


def share_record_batch(record_batch):
    data_size = _get_batch_data_size(record_batch)
    shm = shared_memory.SharedMemory(create=True, size=data_size)
    _write_batch_to_buf(record_batch, pa.py_buffer(shm.buf))

    shm.close()
    # FIXME: we don't need to unregister shared memory and requires this Python process to outlive C++ process if Python fixes it
    # https://bugs.python.org/issue38119
    unregister("/" + shm.name, "shared_memory")
    return shm.name, data_size



from transitions import Machine
from stonewave.sql.udtfs.protocol.fsm.apply.apply_function import ApplyOperator


class ApplyFunctionFsm(object):
    states = ["start", "wait_for_next", "end"]

    def __init__(self, func, batch_sender):
        self.func = func
        self.result = None
        self._batch_sender = batch_sender
        self.machine = Machine(model=self, states=ApplyFunctionFsm.states, initial="start")

        self.machine.add_transition(
            trigger="apply",
            source="start",
            dest="wait_for_next",
            before="apply_params",
        )

        self.machine.add_transition(
            trigger="apply",
            source="wait_for_next",
            dest="wait_for_next",
            before="apply_params",
        )

        self.machine.add_transition(
            trigger="next",
            source="wait_for_next",
            dest="wait_for_next",
            after="send_next_batch",
        )

        self.machine.add_transition(trigger="end", source="*", dest="end", before="end_evaluation")

    def apply_params(self, params, respond):
        apply_op = ApplyOperator(self.func, params)
        self.result = apply_op.execute()
        self.send_next_batch(params, respond)

    def send_next_batch(self, params, respond):
        try:
            batch = next(self.result)
            self._batch_sender.send(batch, respond)
        except StopIteration:
            respond("end")

    def end_evaluation(self, params, respond):
        respond("end")



import pyarrow as pa
from multiprocessing.shared_memory import SharedMemory
from multiprocessing.resource_tracker import unregister
from stonewave.sql.udtfs.logger import logger


class SharedMemoryRecordBatchReader(object):
    shared_memory_objects = []

    def __init__(self, shm_name, shm_size):
        self.shm = SharedMemory(shm_name, create=False, size=shm_size)
        # Adding the shm objects into a class variable essentially makes the shm objects available during the entire lifetime of the process,
        # otherwise, the program still works but Python will complain `BufferError: memoryview has 1 exported buffer`
        # when releasing the shared memory object.
        # The shm objects will be released after the process is terminated, while the buffer in shared memory is managed by the shared memory creator (C++ process)
        SharedMemoryRecordBatchReader.shared_memory_objects.append(self.shm)
        unregister("/" + self.shm.name, "shared_memory")

    def read(self):
        # TODO: figure out if pa.py_buffer(shm.buf) makes any difference
        reader = pa.ipc.open_stream(self.shm.buf)
        batches = list(reader)
        assert len(batches) == 1
        batch = batches[0]
        if batch is not None:
            logger.debug(
                "read record batch via shared memory",
                num_rows=batch.num_rows,
                num_columns=batch.num_columns,
            )
        return batch



import pyarrow as pa
import os
import pandas as pd
from stonewave.sql.udtfs.base_function import BaseFunction, udtf
from stonewave.sql.udtfs.functions.slice_metas.slice_meta_loader import load_stats


class SliceMetasFunction(BaseFunction):
    def __init__(self):
        pass

    def get_name(self):
        return "slice_metas"

    def process(self, row_writer, row_idx, args):
        event_set = args[0]
        home = os.getenv("STONEWAVE_HOME")
        event_set_dir = os.path.join(home, "var", "event_sets", event_set)
        slice_metas = load_stats(event_set_dir)
        for slice_meta in slice_metas:
            slice_meta["event_set"] = event_set

        df = pd.DataFrame(slice_metas)
        table = pa.Table.from_pandas(df, preserve_index=False)
        batches = table.to_batches()
        row_writer.batch_iterator = iter(batches)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version("0.9"):
    raise Exception(
        "Kaitai struct Python API incompatible: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__)
    )


class JavaClass(KaitaiStruct):
    """
    .. seealso::
       Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.1
    """

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.magic = self._io.read_bytes(4)
        if not self.magic == b"\xCA\xFE\xBA\xBE":
            raise kaitaistruct.ValidationNotEqualError(b"\xCA\xFE\xBA\xBE", self.magic, self._io, u"/seq/0")
        self.version_minor = self._io.read_u2be()
        self.version_major = self._io.read_u2be()
        self.constant_pool_count = self._io.read_u2be()
        self.constant_pool = [None] * ((self.constant_pool_count - 1))
        for i in range((self.constant_pool_count - 1)):
            self.constant_pool[i] = JavaClass.ConstantPoolEntry(self._io, self, self._root)

        self.access_flags = self._io.read_u2be()
        self.this_class = self._io.read_u2be()
        self.super_class = self._io.read_u2be()
        self.interfaces_count = self._io.read_u2be()
        self.interfaces = [None] * (self.interfaces_count)
        for i in range(self.interfaces_count):
            self.interfaces[i] = self._io.read_u2be()

        self.fields_count = self._io.read_u2be()
        self.fields = [None] * (self.fields_count)
        for i in range(self.fields_count):
            self.fields[i] = JavaClass.FieldInfo(self._io, self, self._root)

        self.methods_count = self._io.read_u2be()
        self.methods = [None] * (self.methods_count)
        for i in range(self.methods_count):
            self.methods[i] = JavaClass.MethodInfo(self._io, self, self._root)

        self.attributes_count = self._io.read_u2be()
        self.attributes = [None] * (self.attributes_count)
        for i in range(self.attributes_count):
            self.attributes[i] = JavaClass.AttributeInfo(self._io, self, self._root)

    class FloatCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.5
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.value = self._io.read_f4be()

    class AttributeInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.7
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.name_index = self._io.read_u2be()
            self.attribute_length = self._io.read_u4be()
            _on = self.name_as_str
            if _on == u"SourceFile":
                self._raw_info = self._io.read_bytes(self.attribute_length)
                _io__raw_info = KaitaiStream(BytesIO(self._raw_info))
                self.info = JavaClass.AttributeInfo.AttrBodySourceFile(_io__raw_info, self, self._root)
            elif _on == u"LineNumberTable":
                self._raw_info = self._io.read_bytes(self.attribute_length)
                _io__raw_info = KaitaiStream(BytesIO(self._raw_info))
                self.info = JavaClass.AttributeInfo.AttrBodyLineNumberTable(_io__raw_info, self, self._root)
            elif _on == u"Exceptions":
                self._raw_info = self._io.read_bytes(self.attribute_length)
                _io__raw_info = KaitaiStream(BytesIO(self._raw_info))
                self.info = JavaClass.AttributeInfo.AttrBodyExceptions(_io__raw_info, self, self._root)
            elif _on == u"Code":
                self._raw_info = self._io.read_bytes(self.attribute_length)
                _io__raw_info = KaitaiStream(BytesIO(self._raw_info))
                self.info = JavaClass.AttributeInfo.AttrBodyCode(_io__raw_info, self, self._root)
            else:
                self.info = self._io.read_bytes(self.attribute_length)

        class AttrBodyCode(KaitaiStruct):
            """
            .. seealso::
               Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.7.3
            """

            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._read()

            def _read(self):
                self.max_stack = self._io.read_u2be()
                self.max_locals = self._io.read_u2be()
                self.code_length = self._io.read_u4be()
                self.code = self._io.read_bytes(self.code_length)
                self.exception_table_length = self._io.read_u2be()
                self.exception_table = [None] * (self.exception_table_length)
                for i in range(self.exception_table_length):
                    self.exception_table[i] = JavaClass.AttributeInfo.AttrBodyCode.ExceptionEntry(
                        self._io, self, self._root
                    )

                self.attributes_count = self._io.read_u2be()
                self.attributes = [None] * (self.attributes_count)
                for i in range(self.attributes_count):
                    self.attributes[i] = JavaClass.AttributeInfo(self._io, self, self._root)

            class ExceptionEntry(KaitaiStruct):
                """
                .. seealso::
                   Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.7.3
                """

                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._read()

                def _read(self):
                    self.start_pc = self._io.read_u2be()
                    self.end_pc = self._io.read_u2be()
                    self.handler_pc = self._io.read_u2be()
                    self.catch_type = self._io.read_u2be()

                @property
                def catch_exception(self):
                    if hasattr(self, "_m_catch_exception"):
                        return self._m_catch_exception if hasattr(self, "_m_catch_exception") else None

                    if self.catch_type != 0:
                        self._m_catch_exception = self._root.constant_pool[(self.catch_type - 1)]

                    return self._m_catch_exception if hasattr(self, "_m_catch_exception") else None

        class AttrBodyExceptions(KaitaiStruct):
            """
            .. seealso::
               Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.7.5
            """

            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._read()

            def _read(self):
                self.number_of_exceptions = self._io.read_u2be()
                self.exceptions = [None] * (self.number_of_exceptions)
                for i in range(self.number_of_exceptions):
                    self.exceptions[i] = JavaClass.AttributeInfo.AttrBodyExceptions.ExceptionTableEntry(
                        self._io, self, self._root
                    )

            class ExceptionTableEntry(KaitaiStruct):
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._read()

                def _read(self):
                    self.index = self._io.read_u2be()

                @property
                def as_info(self):
                    if hasattr(self, "_m_as_info"):
                        return self._m_as_info if hasattr(self, "_m_as_info") else None

                    self._m_as_info = self._root.constant_pool[(self.index - 1)].cp_info
                    return self._m_as_info if hasattr(self, "_m_as_info") else None

                @property
                def name_as_str(self):
                    if hasattr(self, "_m_name_as_str"):
                        return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

                    self._m_name_as_str = self.as_info.name_as_str
                    return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

        class AttrBodySourceFile(KaitaiStruct):
            """
            .. seealso::
               Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.7.10
            """

            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._read()

            def _read(self):
                self.sourcefile_index = self._io.read_u2be()

            @property
            def sourcefile_as_str(self):
                if hasattr(self, "_m_sourcefile_as_str"):
                    return self._m_sourcefile_as_str if hasattr(self, "_m_sourcefile_as_str") else None

                self._m_sourcefile_as_str = self._root.constant_pool[(self.sourcefile_index - 1)].cp_info.value
                return self._m_sourcefile_as_str if hasattr(self, "_m_sourcefile_as_str") else None

        class AttrBodyLineNumberTable(KaitaiStruct):
            """
            .. seealso::
               Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.7.12
            """

            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._read()

            def _read(self):
                self.line_number_table_length = self._io.read_u2be()
                self.line_number_table = [None] * (self.line_number_table_length)
                for i in range(self.line_number_table_length):
                    self.line_number_table[i] = JavaClass.AttributeInfo.AttrBodyLineNumberTable.LineNumberTableEntry(
                        self._io, self, self._root
                    )

            class LineNumberTableEntry(KaitaiStruct):
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._read()

                def _read(self):
                    self.start_pc = self._io.read_u2be()
                    self.line_number = self._io.read_u2be()

        @property
        def name_as_str(self):
            if hasattr(self, "_m_name_as_str"):
                return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

            self._m_name_as_str = self._root.constant_pool[(self.name_index - 1)].cp_info.value
            return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

    class MethodRefCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.2
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.class_index = self._io.read_u2be()
            self.name_and_type_index = self._io.read_u2be()

        @property
        def class_as_info(self):
            if hasattr(self, "_m_class_as_info"):
                return self._m_class_as_info if hasattr(self, "_m_class_as_info") else None

            self._m_class_as_info = self._root.constant_pool[(self.class_index - 1)].cp_info
            return self._m_class_as_info if hasattr(self, "_m_class_as_info") else None

        @property
        def name_and_type_as_info(self):
            if hasattr(self, "_m_name_and_type_as_info"):
                return self._m_name_and_type_as_info if hasattr(self, "_m_name_and_type_as_info") else None

            self._m_name_and_type_as_info = self._root.constant_pool[(self.name_and_type_index - 1)].cp_info
            return self._m_name_and_type_as_info if hasattr(self, "_m_name_and_type_as_info") else None

    class FieldInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.5
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.access_flags = self._io.read_u2be()
            self.name_index = self._io.read_u2be()
            self.descriptor_index = self._io.read_u2be()
            self.attributes_count = self._io.read_u2be()
            self.attributes = [None] * (self.attributes_count)
            for i in range(self.attributes_count):
                self.attributes[i] = JavaClass.AttributeInfo(self._io, self, self._root)

        @property
        def name_as_str(self):
            if hasattr(self, "_m_name_as_str"):
                return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

            self._m_name_as_str = self._root.constant_pool[(self.name_index - 1)].cp_info.value
            return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

    class DoubleCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.6
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.value = self._io.read_f8be()

    class LongCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.5
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.value = self._io.read_u8be()

    class InvokeDynamicCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.10
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.bootstrap_method_attr_index = self._io.read_u2be()
            self.name_and_type_index = self._io.read_u2be()

    class MethodHandleCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.8
        """

        class ReferenceKindEnum(Enum):
            get_field = 1
            get_static = 2
            put_field = 3
            put_static = 4
            invoke_virtual = 5
            invoke_static = 6
            invoke_special = 7
            new_invoke_special = 8
            invoke_interface = 9

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.reference_kind = KaitaiStream.resolve_enum(
                JavaClass.MethodHandleCpInfo.ReferenceKindEnum, self._io.read_u1()
            )
            self.reference_index = self._io.read_u2be()

    class NameAndTypeCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.6
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.name_index = self._io.read_u2be()
            self.descriptor_index = self._io.read_u2be()

        @property
        def name_as_info(self):
            if hasattr(self, "_m_name_as_info"):
                return self._m_name_as_info if hasattr(self, "_m_name_as_info") else None

            self._m_name_as_info = self._root.constant_pool[(self.name_index - 1)].cp_info
            return self._m_name_as_info if hasattr(self, "_m_name_as_info") else None

        @property
        def name_as_str(self):
            if hasattr(self, "_m_name_as_str"):
                return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

            self._m_name_as_str = self.name_as_info.value
            return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

        @property
        def descriptor_as_info(self):
            if hasattr(self, "_m_descriptor_as_info"):
                return self._m_descriptor_as_info if hasattr(self, "_m_descriptor_as_info") else None

            self._m_descriptor_as_info = self._root.constant_pool[(self.descriptor_index - 1)].cp_info
            return self._m_descriptor_as_info if hasattr(self, "_m_descriptor_as_info") else None

        @property
        def descriptor_as_str(self):
            if hasattr(self, "_m_descriptor_as_str"):
                return self._m_descriptor_as_str if hasattr(self, "_m_descriptor_as_str") else None

            self._m_descriptor_as_str = self.descriptor_as_info.value
            return self._m_descriptor_as_str if hasattr(self, "_m_descriptor_as_str") else None

    class Utf8CpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.7
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.str_len = self._io.read_u2be()
            self.value = (self._io.read_bytes(self.str_len)).decode(u"UTF-8")

    class StringCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.3
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.string_index = self._io.read_u2be()

    class MethodTypeCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.9
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.descriptor_index = self._io.read_u2be()

    class InterfaceMethodRefCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.2
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.class_index = self._io.read_u2be()
            self.name_and_type_index = self._io.read_u2be()

        @property
        def class_as_info(self):
            if hasattr(self, "_m_class_as_info"):
                return self._m_class_as_info if hasattr(self, "_m_class_as_info") else None

            self._m_class_as_info = self._root.constant_pool[(self.class_index - 1)].cp_info
            return self._m_class_as_info if hasattr(self, "_m_class_as_info") else None

        @property
        def name_and_type_as_info(self):
            if hasattr(self, "_m_name_and_type_as_info"):
                return self._m_name_and_type_as_info if hasattr(self, "_m_name_and_type_as_info") else None

            self._m_name_and_type_as_info = self._root.constant_pool[(self.name_and_type_index - 1)].cp_info
            return self._m_name_and_type_as_info if hasattr(self, "_m_name_and_type_as_info") else None

    class ClassCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.1
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.name_index = self._io.read_u2be()

        @property
        def name_as_info(self):
            if hasattr(self, "_m_name_as_info"):
                return self._m_name_as_info if hasattr(self, "_m_name_as_info") else None

            self._m_name_as_info = self._root.constant_pool[(self.name_index - 1)].cp_info
            return self._m_name_as_info if hasattr(self, "_m_name_as_info") else None

        @property
        def name_as_str(self):
            if hasattr(self, "_m_name_as_str"):
                return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

            self._m_name_as_str = self.name_as_info.value
            return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

    class ConstantPoolEntry(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4
        """

        class TagEnum(Enum):
            utf8 = 1
            integer = 3
            float = 4
            long = 5
            double = 6
            class_type = 7
            string = 8
            field_ref = 9
            method_ref = 10
            interface_method_ref = 11
            name_and_type = 12
            method_handle = 15
            method_type = 16
            invoke_dynamic = 18

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.tag = KaitaiStream.resolve_enum(JavaClass.ConstantPoolEntry.TagEnum, self._io.read_u1())
            _on = self.tag
            if _on == JavaClass.ConstantPoolEntry.TagEnum.interface_method_ref:
                self.cp_info = JavaClass.InterfaceMethodRefCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.class_type:
                self.cp_info = JavaClass.ClassCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.utf8:
                self.cp_info = JavaClass.Utf8CpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.method_type:
                self.cp_info = JavaClass.MethodTypeCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.integer:
                self.cp_info = JavaClass.IntegerCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.string:
                self.cp_info = JavaClass.StringCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.float:
                self.cp_info = JavaClass.FloatCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.long:
                self.cp_info = JavaClass.LongCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.method_ref:
                self.cp_info = JavaClass.MethodRefCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.double:
                self.cp_info = JavaClass.DoubleCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.invoke_dynamic:
                self.cp_info = JavaClass.InvokeDynamicCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.field_ref:
                self.cp_info = JavaClass.FieldRefCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.method_handle:
                self.cp_info = JavaClass.MethodHandleCpInfo(self._io, self, self._root)
            elif _on == JavaClass.ConstantPoolEntry.TagEnum.name_and_type:
                self.cp_info = JavaClass.NameAndTypeCpInfo(self._io, self, self._root)

    class MethodInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.6
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.access_flags = self._io.read_u2be()
            self.name_index = self._io.read_u2be()
            self.descriptor_index = self._io.read_u2be()
            self.attributes_count = self._io.read_u2be()
            self.attributes = [None] * (self.attributes_count)
            for i in range(self.attributes_count):
                self.attributes[i] = JavaClass.AttributeInfo(self._io, self, self._root)

        @property
        def name_as_str(self):
            if hasattr(self, "_m_name_as_str"):
                return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

            self._m_name_as_str = self._root.constant_pool[(self.name_index - 1)].cp_info.value
            return self._m_name_as_str if hasattr(self, "_m_name_as_str") else None

    class IntegerCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.4
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.value = self._io.read_u4be()

    class FieldRefCpInfo(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.4.2
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.class_index = self._io.read_u2be()
            self.name_and_type_index = self._io.read_u2be()

        @property
        def class_as_info(self):
            if hasattr(self, "_m_class_as_info"):
                return self._m_class_as_info if hasattr(self, "_m_class_as_info") else None

            self._m_class_as_info = self._root.constant_pool[(self.class_index - 1)].cp_info
            return self._m_class_as_info if hasattr(self, "_m_class_as_info") else None

        @property
        def name_and_type_as_info(self):
            if hasattr(self, "_m_name_and_type_as_info"):
                return self._m_name_and_type_as_info if hasattr(self, "_m_name_and_type_as_info") else None

            self._m_name_and_type_as_info = self._root.constant_pool[(self.name_and_type_index - 1)].cp_info
            return self._m_name_and_type_as_info if hasattr(self, "_m_name_and_type_as_info") else None



from stonewave.sql.udtfs.constants import VALIDATE_ERROR_MSG_TEMPLATE
from stonewave.sql.udtfs.constants import ParameterDataType as pdt
from stonewave.sql.udtfs.constants import ParameterType as pt
from stonewave.sql.udtfs.table_row_writer import TableRowWriter
from stonewave.sql.udtfs.protocol.ipc import share_record_batch
from stonewave.sql.udtfs.protocol.fsm.apply.apply_function import ApplyOperator
import pyarrow as pa
import pyarrow.csv as csv


def _get_expected_python_type_for_parameter_type(parameter_type):
    if parameter_type == pdt.STRING:
        return str
    if parameter_type == pdt.INT:
        return int
    if parameter_type == pdt.FLOAT:
        return float
    if parameter_type == pdt.BOOL:
        return bool


def _validate_apply_parameters(given_params, expected_parameters_list):
    """
    Validate apply table function parameter type.
    :param given_params example list [{"column_name": "column"}, {"a": "literal"}]
    :expected_parameters_list all possible parameter lists
    Will not check the column datatype, because column datatype can be auto casted by stonewave.
    Will not check the size of parameter, only check the existing parameter datatype.
    Allow default parameter for table function
    """
    for expected_params in expected_parameters_list:
        if len(given_params) == len(expected_params):
            for i in range(len(given_params)):
                if given_params[i][pt.TYPE] == pt.LITERAL:
                    expected_type = expected_params[i].split(" = ")[0].upper()
                    if not isinstance(
                        given_params[i][pt.VALUE], _get_expected_python_type_for_parameter_type(expected_type)
                    ):
                        raise Exception(VALIDATE_ERROR_MSG_TEMPLATE.format(",".join(expected_params)))
            return
    raise Exception(VALIDATE_ERROR_MSG_TEMPLATE.format(",".join(expected_parameters_list)))


def _validate_eval_parameters(given_params, expected_parameters_list):
    """
    Validate evaluate table function parameter type.
    :param given_params example list ['abc', 123]
    :expected_parameters_list all possible parameter lists
    """
    for expected_params in expected_parameters_list:
        if len(given_params) == len(expected_params):
            for i in range(len(given_params)):
                expected_type = expected_params[i].split(" = ")[0].upper()
                if not isinstance(given_params[i], _get_expected_python_type_for_parameter_type(expected_type)):
                    raise Exception(VALIDATE_ERROR_MSG_TEMPLATE.format(",".join(expected_params)))
            return
    raise Exception(VALIDATE_ERROR_MSG_TEMPLATE.format(",".join(expected_parameters_list)))


def wrap_apply_function_input_parameter_with_type(parameter_type, parameter_value):
    """
    Wrap apply function input parameter with type.
    :param parameter_type support [stonewave.sql.udtfs.constants.ParameterType.COLUMN,
           stonewave.sql.udtfs.constants.ParameterType.LITERAL]
    :param parameter_value
    :return:
    """
    return {pt.TYPE: parameter_type, pt.VALUE: parameter_value}


def check_expected_parameters_list(expected_parameters_list):
    parameters_size = set()
    for expected_params in expected_parameters_list:
        params_len = len(expected_params)
        if params_len in parameters_size:
            raise Exception(
                "Do not support same parameter list size with different datatype, because of \
                Stonewave can auto cast"
            )
        parameters_size.add(params_len)
        for i in range(len(expected_params)):
            expected_type = expected_params[i].split(" = ")[0].upper()
            if expected_type not in [pdt.STRING, pdt.INT, pdt.FLOAT, pdt.BOOL, pdt.TABLE]:
                raise Exception("Parameter datatype should be in one of [STRING, INT, FLOAT, BOOL, TABLE]")


def eval_table_function_test(func_method, params, expected_parameters_list):
    """
    Test for evaluate table function
    :param func_method table function method
    :param params evaluate table function input parameters, only literal supported
    :param expected_parameters_list expected parameters list
    """
    _validate_eval_parameters(params, expected_parameters_list)
    func = func_method()
    row_writer = TableRowWriter()
    func.process(row_writer, 0, params)
    batch = None
    begininig = True
    while begininig or batch is None:
        begininig = False
        batch = row_writer.flush(True)
        yield batch


def apply_table_function_test(func_method, left_table, is_outer_apply, params, expected_parameters_list):
    _validate_apply_parameters(params, expected_parameters_list)
    func = func_method()
    shm_info = share_record_batch(left_table)
    operator_params = [{"shm_name": shm_info[0], "shm_size": shm_info[1]}, is_outer_apply, ""]
    operator_params.extend(params)

    apply_op = ApplyOperator(func, operator_params)
    result = apply_op.execute()
    return result


def check_batch_iterator_results_with_csv_file(batch_iterator, expected_result_csv_path):
    """
    make batch iterator to table and compared with csv file table
    :param
    """
    batches = []
    try:
        batches.append(next(batch_iterator))
    except StopIteration:
        pass
    table = pa.Table.from_batches(batches)
    # empty string cell will be read as null cell
    convert_options = csv.ConvertOptions(strings_can_be_null=True)
    expected_table = csv.read_csv(expected_result_csv_path, convert_options=convert_options)
    assert table.num_rows == expected_table.num_rows
    assert table.num_columns == expected_table.num_columns
    assert table.equals(expected_table)


def read_arrow_batch_from_csv(csv_pth):
    # empty string cell will be read as null cell
    convert_options = csv.ConvertOptions(strings_can_be_null=True)
    arrow_table = csv.read_csv(csv_pth, convert_options=convert_options)
    return arrow_table.to_batches()[0]


# TODO: add test utility for table-valued parameter / default value



import pyarrow as pa
from stonewave.sql.udtfs.constants import ParameterDataType as pdt


def get_arrow_data_type_from_value(value):
    if not value:
        return pa.utf8()
    else:
        # bool needs to be placed before int because isinstance(True, int) == True
        if isinstance(value, bool):
            return pa.bool_()
        elif isinstance(value, int):
            return pa.int64()
        elif isinstance(value, float):
            return pa.float64()
        # elif isinstance(value, decimal.Decimal):
        # return pa.decimal128(13)
        else:
            return pa.utf8()


def get_arrow_data_type(type):
    if type == pdt.STRING:
        return pa.utf8()
    elif type == pdt.FLOAT:
        return pa.float64()
    elif type == pdt.INT:
        return pa.int64()
    elif type == pdt.BOOL:
        return pa.bool_()
    else:
        return pa.utf8()



from stonewave.sql.udtfs.table_row_writer import TableRowWriter
from stonewave.sql.udtfs.protocol.fsm.apply import _lateral_join, _get_row


def _equal_length_apply(func, left_table, column_name_qualifier, func_params):
    # this can only be used with outer apply
    # only table function writing results row by row with a batch builder can correctly function here
    right_table_writer = TableRowWriter()
    func.initialize(right_table_writer)

    right_table = None
    for i in range(0, left_table.num_rows):
        row = _get_row(left_table, i, func_params)
        try:
            func.process(right_table_writer, i, row)
            if right_table_writer.builder_row_count() <= i:
                right_table_writer.record_batch_builder.append_null_row()
        except Exception as e:
            right_table_writer.record_batch_builder.append_null_row()
    right_table = right_table_writer.flush(forced=True)
    joined_table = _lateral_join(left_table, right_table, column_name_qualifier)
    return joined_table



from stonewave.sql.udtfs.record_batch_builder import RecordBatchBuilder


class ApplyLeftTableBuilder(RecordBatchBuilder):
    def __init__(self, left_table):
        super().__init__()
        self.left_table = left_table
        self.left_table_field_names = left_table.schema.names
        for name, type in zip(self.left_table_field_names, left_table.schema.types):
            self.add_column(name, type)
        self.total_repeats = 0

    def repeat(self, row_idx, repeats):
        # repeat left_table[row_idx] multiple times, determined by `repeats` in the batch builder
        if repeats > 0:
            self.total_repeats += repeats
            for fi in range(0, len(self.left_table_field_names)):
                cell_value = self.left_table.column(fi)[row_idx].as_py()
                for i in range(repeats):
                    self.append(self.left_table_field_names[fi], cell_value)
            self.increase_row_count(repeats)

    def get_total_repeats(self):
        return self.total_repeats



from transitions import Machine
from stonewave.sql.udtfs.protocol import ipc
from stonewave.sql.udtfs.protocol.fsm.shared_record_batch_reader import read_record_batch
from stonewave.sql.udtfs.table_row_writer import TableRowWriter
from stonewave.sql.udtfs.constants import PROCESS_ERROR_MSG_TEMPLATE


class EvalFunctionWithTableParamFsm(object):
    states = ["start", "wait_for_next_table_batch", "wait_for_next", "end"]

    def __init__(self, func, batch_sender):
        self.func = func
        self._batch_sender = batch_sender
        self.machine = Machine(model=self, states=EvalFunctionWithTableParamFsm.states, initial="start")

        self.machine.add_transition(
            trigger="eval_with_table_param",
            source="start",
            dest="wait_for_next_table_batch",
            after="eval_params",
        )

        self.machine.add_transition(
            trigger="eval_with_table_param",
            source="wait_for_next_table_batch",
            dest="wait_for_next_table_batch",
            after="eval_params",
        )

        self.machine.add_transition(
            trigger="end_table_param",
            source="wait_for_next_table_batch",
            dest="wait_for_next",
            after="send_next_batch",
        )

        self.machine.add_transition(
            trigger="next",
            source="wait_for_next",
            dest="wait_for_next",
            after="send_next_batch",
        )

        self.machine.add_transition(trigger="end", source="wait_for_next", dest="end", before="end_evaluation")

        self.machine.add_transition(trigger="*", source="end", dest="end", before="end_evaluation")
        self._row_writer = TableRowWriter()
        self.func.initialize(self._row_writer)
        self._batches = []
        self._batch_idx = 0
        self._read_record_batch_func = read_record_batch

    @property
    def read_record_batch_func(self):
        return self._read_record_batch_func

    # a method used for injecting a different function for unit testing purpose
    @read_record_batch_func.setter
    def read_record_batch_func(self, value):
        self._read_record_batch_func = value

    def eval_params(self, params, respond):
        batch = self._read_record_batch_func(params[0])

        params[0] = batch
        try:
            func_name = self.func.get_name()
            if batch is not None:
                self.func.process(self._row_writer, 0, params)
                next_batch = self._row_writer.flush(forced=False)
                if next_batch:
                    self._batches.append(next_batch)
                respond("next_table_batch")
            else:
                self.func.process(self._row_writer, 0, params)
                self.end_table_param(params, respond)
        except Exception as e:
            raise Exception(PROCESS_ERROR_MSG_TEMPLATE.format(func_name, str(e)))

    def send_next_batch(self, params, respond):
        if self._batch_idx < len(self._batches):
            batch = self._batches[self._batch_idx]
            self._batch_idx += 1
            self._batch_sender.send(batch, respond)
            return
        batch = self._row_writer.flush()
        if batch is not None:
            self._batch_sender.send(batch, respond)
            return
        batch = self._row_writer.flush(forced=True)
        if batch is not None:
            self._batch_sender.send(batch, respond)
            return
        self.end(params, respond)

    def end_evaluation(self, params, respond):
        respond("end")



from faker import Faker
import pyarrow as pa
from faker_web import WebProvider
from faker_vehicle import VehicleProvider
from faker_wifi_essid import WifiESSID as WifiEssidProvider
from faker_microservice import Provider as MicroServiceProvider
from faker_credit_score import CreditScore as CreditScoreProvider
import decimal
from stonewave.sql.udtfs.base_function import BaseFunction
from stonewave.sql.udtfs.logger import logger

fake = Faker()
fake.add_provider(WebProvider)
fake.add_provider(WifiEssidProvider)
fake.add_provider(MicroServiceProvider)
fake.add_provider(VehicleProvider)
fake.add_provider(CreditScoreProvider)


def _get_data_type(field_faker):
    if not field_faker:
        return pa.utf8()
    else:
        value = field_faker()
        # bool needs to be placed before int because isinstance(True, int) == True
        if isinstance(value, bool):
            return pa.bool_()
        elif isinstance(value, int):
            return pa.int64()
        elif isinstance(value, float):
            return pa.float64()
        # elif isinstance(value, decimal.Decimal):
        # return pa.decimal128(13)
        else:
            return pa.utf8()


class FakerFunction(BaseFunction):
    def __init__(self):
        pass

    def get_name(self):
        return "faker"

    def _safe_cast(self, value, to_type, default=None):
        try:
            return to_type(value)
        except (ValueError, TypeError):
            return default

    def process(self, row_writer, row_idx, args):
        row_count = max(0, self._safe_cast(args[0], int, 0))
        if args[1] is None:
            logger.debug("faker column name is not provided, no row is produced")
            return
        columns = [c.strip() for c in args[1].split(",")]
        non_duplicated_column_names = self._non_duplicated_column_names(columns)

        record_batch_builder = row_writer.record_batch_builder
        for fake_column_method, fake_column_name in zip(columns, non_duplicated_column_names):
            field_faker = getattr(fake, fake_column_method, None)
            field_type = _get_data_type(field_faker)

            record_batch_builder.add_column(fake_column_name, field_type)

            field_faker = getattr(fake, fake_column_method, None)
            array = [field_faker() if field_faker else None for i in range(row_count)]
            record_batch_builder.extend(fake_column_name, array)
        record_batch_builder.increase_row_count(row_count)
        logger.debug(
            "rows generated via faker",
            row_count=row_count,
            builder_row_count=record_batch_builder.row_count(),
        )

    def _non_duplicated_column_names(self, columns):
        column_set = set()
        non_duplicated_column_names = []
        for column in columns:
            column_id = 0
            non_duplicated_column_name = column
            if non_duplicated_column_name in column_set:
                column_id += 1
                non_duplicated_column_name = "{}${}".format(column, column_id)
            column_set.add(non_duplicated_column_name)
            non_duplicated_column_names.append(non_duplicated_column_name)
        return non_duplicated_column_names



import pyarrow as pa
from stonewave.sql.udtfs.base_function import BaseFunction
from stonewave.sql.udtfs.logger import logger


class TransposeFunction(BaseFunction):
    def __init__(self):
        self.tables = []
        self.header_field = None

    def get_name(self):
        return "transpose"

    def process(self, row_writer, row_idx, args):
        assert len(args) > 0
        batch = args[0]
        if batch is not None:
            if not self.header_field and len(args) == 2:
                self.header_field = args[1]
            table = pa.Table.from_batches([batch])
            self.tables.append(table)
        else:
            self.transpose(row_writer)

    def transpose(self, row_writer):
        if self.tables:
            table = pa.concat_tables(self.tables, promote=True)
            df = table.to_pandas()
            header_field_index = None
            # before transpose, remove duplicate header field value
            if self.header_field:
                if self.header_field not in df.columns:
                    raise Exception("Table function 'TRANSPOSE' parameter invalid: " "no such header field")
                header_field_index = df.columns.get_loc(self.header_field)
                df = df.drop_duplicates(subset=[self.header_field])
                logger.debug(
                    "remove duplicate field values row", before_num_rows=table.num_rows, after_num_rows=len(df)
                )

            # transpose
            transpose_df = df.transpose()
            transpose_df = transpose_df.reset_index()
            # find header field row index
            if header_field_index is not None:
                # promote header field as header
                new_header = transpose_df.iloc[header_field_index]
                transpose_df = transpose_df.drop(transpose_df.index[header_field_index])
                try:
                    # fill field header null value with string "NULL"
                    transpose_df.columns = new_header.fillna("NULL")
                except Exception as e:
                    raise Exception("Table function 'TRANSPOSE' parameter invalid: " "unsupported header field, %s" % e)
            else:
                transpose_df.columns = ["col_%d" % i for i in range(transpose_df.columns.size)]

            try:
                table = pa.Table.from_pandas(transpose_df, preserve_index=False)
            except pa.lib.ArrowTypeError as e:
                raise Exception("Failed to transpose: " "there are multiple datatype in single field, %s" % str(e))
            except Exception as e:
                raise Exception("Failed to transpose: %s" % e)
            batches = table.to_batches()
            row_writer.batch_iterator = iter(batches)
        else:
            return



import json
import datetime
from functools import cmp_to_key
import os


def _format_epoch(timestamp):
    return datetime.datetime.fromtimestamp(timestamp / 1000000).strftime("%Y-%m-%d %H:%M:%S")


def _compare_slice(s1, s2):
    if s1["latest"] > s2["latest"] or s1["latest"] == s2["latest"] and s1["earliest"] > s2["earliest"]:
        return -1
    else:
        return 1


def _parse_slice_meta(slice_meta_file):
    with open(slice_meta_file, "r") as meta:
        slice_meta = meta.read()
        slices = json.loads(slice_meta)
        for slice in slices:
            slice["earliest"] = _format_epoch(slice["earliest_time"])
            slice["latest"] = _format_epoch(slice["latest_time"])
        return slices


def _add_slice_meta(all_metas, slice_meta_file, sub_dir=""):
    if os.path.exists(slice_meta_file):
        slices = _parse_slice_meta(slice_meta_file)
        if sub_dir:
            for slice in slices:
                slice["partition"] = sub_dir
        all_metas.extend(slices)


def load_stats(event_set_dir):
    sorted_slices = []
    if os.path.exists(event_set_dir):
        all_metas = []
        dirs = os.listdir(event_set_dir)
        for sub_dir in dirs:
            slice_meta_json = os.path.join(event_set_dir, sub_dir, "slice_meta.json")
            _add_slice_meta(all_metas, slice_meta_json, sub_dir)

        # the new version stores slice_meta.json in different location
        slice_meta_json = os.path.join(event_set_dir, "slice_meta.json")
        _add_slice_meta(all_metas, slice_meta_json)
        sorted_slices = sorted(all_metas, key=cmp_to_key(_compare_slice))
    return sorted_slices
